"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family line pattern data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities import utility as util
from duHast.Revit.Categories import categories as rCats
from duHast.Revit.LinePattern.line_patterns import (
    get_line_pattern_from_category,
    get_line_pattern_from_level_element,
    PROPERTY_PATTERN_ID,
)
from duHast.Revit.Levels.levels import get_levels_list_ascending


# import Autodesk
import Autodesk.Revit.DB as rdb

# data dictionary key values specific to this class
PATTERN_NAME = "patternName"
PATTERN_ID = "patternId"


class LinePatternData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None, data_type=None):
        """
        Class constructor

        :param rootPath: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param dataType: Human readable data type descriptor
        :type dataType: str
        """

        super(LinePatternData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
            data_type=data_type,
        )
        # super(CategoryData, self).__init__(rootPath, dataType)
        """
        self.data = []
        
        if(dataType != None):
            self.dataType = dataType
        else:
            self.dataType = 'not declared'
        
        if(rootPath != None):
            self.rootPath = rootPath
        else:
            self.rootPath = '-'

        if(rootCategoryPath != None):
            self.rootCategoryPath = rootCategoryPath
        else:
            self.rootCategoryPath = '-'
        """

    def _add_category_to_dic(self, line_pattern_ids, pattern_id, category):
        """
        Adds a category to a dictionary using the line pattern ID as value

        :param linePatternIds: dictionary in format: Key: line Pattern Id, value list of categories using that line pattern
        :type linePatternIds: {ElementId:[Autodesk.revit.DB.Category]}
        :param patternId: the pattern element id
        :type patternId: Autodesk.Revit.DB.ElementId
        :param category: the category using the line pattern
        :type category: Autodesk.Revit.DB.Category
        """

        if pattern_id in line_pattern_ids:
            line_pattern_ids[pattern_id].append(category)
        else:
            line_pattern_ids[pattern_id] = [category]

    def _get_line_pattern_from_categories(self, doc):
        """
        Loops over all family categories and sub categories and any other categories and sub categories with a positive Id (custom sub category) and
        returns a dictionary of line pattern ids to categories using them.

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        :return: a dictionary in format: Key: line Pattern Id, value list of categories using that line pattern
        :rtype: {ElementId:[Autodesk.revit.DB.Category]}
        """

        line_pattern_ids_to_categories = {}

        # get any line pattern added to the family category itself
        main_cat = rCats.get_family_category(doc)
        for m_cat_name in main_cat:
            l_style = get_line_pattern_from_category(main_cat[m_cat_name], doc)
            # update dictionary
            self._add_category_to_dic(
                line_pattern_ids_to_categories,
                l_style[PROPERTY_PATTERN_ID],
                main_cat[m_cat_name],
            )

        # get line patterns from sub categories of the family category
        main_cats = rCats.get_main_sub_categories(doc)
        for m_cat_name in main_cats:
            l_style = get_line_pattern_from_category(main_cats[m_cat_name], doc)
            # update dictionary
            self._add_category_to_dic(
                line_pattern_ids_to_categories,
                l_style[PROPERTY_PATTERN_ID],
                main_cats[m_cat_name],
            )

        # get line pattern from unrelated sub categories
        sub_cats_other = rCats.get_other_sub_categories(doc)
        for s_cat_name in sub_cats_other:
            for s_cat_item in sub_cats_other[s_cat_name]:
                # only use custom categories not build in ones (id smaller then 0)
                if sub_cats_other[s_cat_name][s_cat_item].Id.IntegerValue > 0:
                    l_style = get_line_pattern_from_category(
                        sub_cats_other[s_cat_name][s_cat_item], doc
                    )
                    # update dictionary
                    self._add_category_to_dic(
                        line_pattern_ids_to_categories,
                        l_style[PROPERTY_PATTERN_ID],
                        sub_cats_other[s_cat_name][s_cat_item],
                    )

        # get line pattern from reference lines and planes categories import in families main cat
        other_cats = rCats.get_category_by_built_in_def_name(
            doc,
            [
                rdb.BuiltInCategory.OST_ReferenceLines,  # reference lines
                rdb.BuiltInCategory.OST_CLines,  # reference planes
                rdb.BuiltInCategory.OST_ImportObjectStyles,
            ],  # import in families
        )
        for o_cat in other_cats:
            l_style = get_line_pattern_from_category(o_cat, doc)
            # update dictionary
            self._add_category_to_dic(
                line_pattern_ids_to_categories, l_style[PROPERTY_PATTERN_ID], o_cat
            )

        return line_pattern_ids_to_categories

    def _get_pattern_from_level_element(self, doc):
        """
        Gets the pattern data from all level types in document.

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document

        :return: a dictionary in format: Key: line Pattern Id, value list of levels using that line pattern
        :rtype: {ElementId:[Autodesk.revit.DB.Category]}
        """

        level_pattern_data = {}
        levels = get_levels_list_ascending(doc)
        for level in levels:
            pattern_data = get_line_pattern_from_level_element(doc, level)
            self._add_category_to_dic(
                level_pattern_data, pattern_data[PROPERTY_PATTERN_ID], level
            )
        return level_pattern_data

    def _get_pattern_name(self, element):
        """
        Get the element name.

        :param element: _description_
        :type element: _type_
        :return: Element name, or exception stating name is not unicode
        :rtype: str
        """

        element_name = "unknown_notUnicode"
        try:
            element_name = util.encode_ascii(rdb.Element.Name.GetValue(element))
        except Exception as ex:
            element_name = element_name + " Exception: " + str(ex)
        return element_name

    def _get_pattern_usage_data_from_categories(self, line_pattern_ids, element):
        """
        Returns how often and on which category a line pattern is used

        :param linePatternIds: a dictionary in format: Key: line Pattern Id, value list of categories using that line pattern
        :type linePatternIds: {ElementId:[Autodesk.revit.DB.Category]}
        :param element: The line pattern element.
        :type element: _type_

        :return: counter and  a list of dictionaries in format {categoryId: int, categoryName: str}
        :rtype: int, [{categoryId:int, categoryName:str}]
        """

        counter = 0
        pattern_names = []
        # how often used
        if element.Id in line_pattern_ids:
            counter = len(line_pattern_ids[element.Id])
            for pat in line_pattern_ids[element.Id]:
                pattern_names.append(
                    {"categoryId": pat.Id.IntegerValue, "categoryName": pat.Name}
                )
        return counter, pattern_names

    def _get_pattern_usage_data_from_level(self, line_pattern_ids, element):
        """
        Returns how often and on which Level type a line pattern is used

        :param linePatternIds:
        :type linePatternIds: {ElementId:[Autodesk.revit.DB.Category]}
        :param element: The line pattern element.
        :type element: _type_

        :return: counter and  a list of dictionaries in format {categoryId: int, categoryName: str}
        :rtype: int, [{categoryId:int, categoryName:str}]
        """

        counter = 0
        pattern_names = []

        # how often used
        if element.Id in line_pattern_ids:
            counter = len(line_pattern_ids[element.Id])
            for pat in line_pattern_ids[element.Id]:
                pattern_names.append(
                    {
                        "levelId": pat.Id.IntegerValue,
                        "levelTypeName": rdb.Element.Name.GetValue(pat),
                    }
                )
        return counter, pattern_names

    def process(self, doc):
        """
        Collects all line pattern data from the document and stores it in the class property .data

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # get all line patterns used in categories (includes sub categories of family category and any custom subcategories of non family category present, includes also
        # ref planes , ref lines, import styles)
        line_pattern_ids_by_category = self._get_line_pattern_from_categories(doc)
        # get line pattern used on level element
        line_pattern_ids_by_from_level = self._get_pattern_from_level_element(doc)

        collector = rdb.FilteredElementCollector(doc).OfClass(rdb.LinePatternElement)
        for element in collector:
            # just in case parameter name is not unicode
            element_name = self._get_pattern_name(element)
            # get usage data from categories
            counter, pattern_names = self._get_pattern_usage_data_from_categories(
                line_pattern_ids_by_category, element
            )
            # get usage data from levels
            (
                counter_level,
                pattern_names_level,
            ) = self._get_pattern_usage_data_from_level(
                line_pattern_ids_by_from_level, element
            )

            # get overall count
            counter = counter + counter_level
            # get overall usage data
            usage_all = pattern_names + pattern_names_level

            # build data
            self.data.append(
                {
                    IFamData.ROOT: self.root_path,
                    IFamData.ROOT_CATEGORY: self.root_category_path,
                    IFamData.FAMILY_NAME: self._strip_file_extension(doc.Title),
                    IFamData.FAMILY_FILE_PATH: doc.PathName,
                    IFamData.USAGE_COUNTER: counter,
                    IFamData.USED_BY: usage_all,
                    PATTERN_NAME: element_name,
                    PATTERN_ID: element.Id.IntegerValue,
                }
            )

    def get_data(self):
        return self.data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to rename family files on a local or network drive.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This helper function expect a folder containing rename directive files. For format of those files refer to module RevitFamilyRenameFilesUtils

Note:

- The revit category is not used when renaming files but when renaming nested families.
- Any associated type catalogue files will also be renamed to match the new family name.
- Rename directives may not have the filePath property set if the directive is only meant to be used on loaded families.

"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import clr
# import System
import os

from duHast.Revit.Family import family_rename_files_utils as rFamRenameUtils
from duHast.Utilities.Objects import result as res
from duHast.Utilities import files_io as fileIO


def _rename_files(rename_directives):
    """
    Renames family files and any associated catalogue files based on rename directives.

    :param rename_directives: List of tuples representing rename directives.
    :type rename_directives: [rename_directive]

    :return:
        Result class instance.

        - result.status. True if files where renamed successfully, otherwise False.
        - result.message will contain each rename messages in format 'old name -> new name'.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    return_value.update_sep(True, "Renaming families:")

    for rename_directive in rename_directives:
        try:
            return_value.append_message("rename directive: {}".format(rename_directive))
            # check if rename directive includes a file path ( might be empty if nested families only are to be renamed)
            if rename_directive.filePath != "":
                # attempt to rename family file
                try:
                    # build the new file name
                    new_full_name = os.path.join(
                        os.path.dirname(rename_directive.filePath),
                        rename_directive.newName + ".rfa",
                    )
                    if fileIO.file_exist(rename_directive.filePath):
                        os.rename(rename_directive.filePath, new_full_name)
                        return_value.append_message(
                            "{} -> {}".format(
                                rename_directive.name, rename_directive.newName
                            )
                        )
                    else:
                        return_value.update_sep(
                            False, "File not found: {}".format(rename_directive.name)
                        )
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "Failed to rename file: {} with exception: {}".format(
                            rename_directive.name, e
                        ),
                    )

                # take care of catalogue files as well
                old_full_name = rename_directive.filePath[:-4] + ".txt"
                new_full_name = os.path.join(
                    os.path.dirname(rename_directive.filePath),
                    rename_directive.newName + ".txt",
                )
                oldname = rename_directive.name + ".txt"
                newname = rename_directive.newName + ".txt"
                try:
                    if fileIO.file_exist(old_full_name):
                        os.rename(old_full_name, new_full_name)
                        return_value.append_message("{} -> {}".format(oldname, newname))
                    else:
                        return_value.update_sep(
                            True, "No catalogue file found: {}".format(oldname)
                        )  # nothing gone wrong here...just no catalogue file present
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "Failed to rename file: {} with exception: {}".format(
                            old_full_name, e
                        ),
                    )
            else:
                return_value.update_sep(
                    True, "No file path found: {}".format(rename_directive.name)
                )  # nothing gone wrong here...just not required to rename a file
        except Exception as e:
            return_value.update_sep(
                False, "Failed to rename files with exception: ".format(e)
            )
    return return_value


def rename_family_files(directory_path):
    """
    Entry point for this module. Will read rename directives files in given directory and attempt to rename
    family files and any associated catalogue files accordingly.

    Note: Rename directive may not include a file path in situations where a loaded family only is to be renamed. This \
        will still return True in such a case.

    :param directory_path: Fully qualified directory path to where rename directive files are located.
    :type directory_path: str
    :return: 
        Result class instance.

        - result.status. True if files where renamed successfully, otherwise False.
        - result.message will contain each rename messages in format 'old name -> new name'.
        - result.result empty list
        
        On exception:
        
        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get directives from folder
    rename_directives_result = rFamRenameUtils.get_rename_directives(directory_path)
    # check if anything came back
    if rename_directives_result.status:
        rename_directives = rename_directives_result.result
        # rename files as per directives
        return_value = _rename_files(rename_directives)
    else:
        return_value = rename_directives_result

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Base class for objects.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This class provides some utility functions to all child classes:

- __repr__() a way of providing detailed debug output through print
- to_json() A json formatted dump of the class

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json


"""
The `Base` class is a parent class that provides common functionalities and methods for its subclasses. It includes a constructor, a debug output method, a comparison method, a hash method, a method to convert the instance to JSON, a method to convert string properties to UTF-8 in JSON conversion, a method to check if an object is a Python primitive, and a method to convert the class to a dictionary.

Example Usage:
    # Creating a subclass of Base
    class MyClass(Base):
        def __init__(self, name):
            super(MyClass, self).__init__()
            self.name = name

    # Creating an instance of MyClass
    obj = MyClass("example")

    # Printing the debug output
    print(obj)

    # Comparing two instances
    obj1 = MyClass("example")
    obj2 = MyClass("example")
    print(obj1 == obj2)

    # Converting the instance to JSON
    json_data = obj.to_json()

    # Converting the instance to JSON with UTF-8 string properties
    json_data_utf = obj.to_json_utf()

    # Converting the instance to a dictionary
    dict_data = obj.class_to_dict()

Main functionalities:
- The `Base` class allows for multi-inheritance in its subclasses.
- It provides a debug output method that returns a string representation of the class properties.
- It provides a comparison method that checks if two instances are of the same class.
- It provides a hash method required for the custom comparison method.
- It provides a method to convert the instance to JSON.
- It provides a method to convert string properties to UTF-8 in JSON conversion.
- It provides a method to check if an object is a Python primitive.
- It provides a method to convert the class to a dictionary.

Methods:
- `__init__(self, **kwargs)`: The class constructor that forwards all unused arguments to the super class.
- `__repr__(self)`: Enables detailed debug output of all class properties.
- `__eq__(self, other)`: Custom compare is equal override.
- `__hash__(self)`: Custom hash override.
- `to_json(self)`: Convert the instance of this class to JSON.
- `string_to_utf(self, o)`: Convert any properties stored as string to UTF-8 in JSON conversion.
- `to_json_utf(self)`: Convert the instance of this class to JSON with UTF-8 string properties.
- `_is_primitive(self, obj)`: Checks whether an object is a Python primitive.
- `class_to_dict(self)`: Returns all class properties and their values as a dictionary.

Fields:
- No specific fields are defined in the `Base` class.
"""
class Base(object):
    def __init__(self, **kwargs):
        """
        Class constructor
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!

        super(Base, self).__init__(**kwargs)

    def __repr__(self):
        """
        Enables detailed debug output of all class properties using: rep(obj)

        :return: A string listing class properties and their respective values.
        :rtype: string
        """

        return "{}({})".format(
            self.__class__.__name__,
            ", ".join("{}={!r}".format(k, v) for k, v in self.__dict__.items()),
        )


    def __eq__(self, other):
        """
        Custom compare is equal override 

        :param other: Another instance of pattern class
        :type other: :class:`.PatternBase`
        :return: True if name value of other colour class instance equal the name values of this instance, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, self.__class__)

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)
    
    def __hash__(self):
        """
        Custom hash override

        Required due to custom __eq__ override present in this class
        """
        try:
            return hash(self.__class__)
        except Exception as e:
            raise ValueError(
                "Exception {} occurred in {} with values: name:{}".format(
                    e, self.data_type, self.name
                )
            )


    def to_json(self):
        """
        Convert the instance of this class to json.

        :return: A Json object.
        :rtype: json
        """

        return json.dumps(self, indent=None, default=lambda o: o.__dict__)


    def string_to_utf(self, o):
        '''
        Used to convert any properties stored as string to utf-8 in to json conversion of all class properties...

        :param o: _description_
        :type o: _type_
        :return: _description_
        :rtype: _type_
        '''

        if isinstance(o, str):
            return o.encode('utf-8').decode('utf-8')  # Encoding and decoding to ensure the type is str
        return o.__dict__

    def to_json_utf(self):
        '''
        Convert the instance of this class to json, any string properties are converted to utf-8

        :return: A Json object.
        :rtype: json
        '''

        return json.dumps(self, indent=None, default=self.string_to_utf, ensure_ascii=False)



    def _is_primitive(self, obj):
        """
        Checks whether object past in is a python primitive

        :param obj: The object to be tested.
        :type obj: obj
        :return: True if object is a python primitive, Otherwise False.
        :rtype: Bool
        """

        return isinstance(obj, (int, float, str, bool))

    def class_to_dict(self):
        """
        Returns all class properties and their values as a dictionary

        :return: A dictionary of all class properties names and their values
        :rtype: {str:var,}
        """

        if isinstance(self, object):
            class_dict = {}
            for key, value in self.__dict__.items():
                if self._is_primitive(value):
                    class_dict[key] = value
                else:
                    class_dict[key] = value.class_to_dict()
            return class_dict
        else:
            return self



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family base data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities import directory_io as dirIO, files_io as fileIO
from duHast.Revit.Common import file_io as rFile
from duHast.Revit.Family.Data import family_base_data_utils as rFamBaseDataUtils

# import Autodesk
# import Autodesk.Revit.DB as rdb

# data dictionary key values specific to this class
CATEGORY_NAME = "categoryName"


class FamilyBaseData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None, data_type=None):
        """
        Class constructor

        :param rootPath: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param dataType: Human readable data type descriptor
        :type dataType: str
        """

        # store data type  in base class
        super(FamilyBaseData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
            data_type=data_type,
        )
        # super(CategoryData, self).__init__(rootPath, dataType)

        """
        self.data = []
        
        if(dataType != None):
            self.dataType = dataType
        else:
            self.dataType = 'not declared'
        
        if(rootPath != None):
            self.rootPath = rootPath
        else:
            self.rootPath = '-'

        if(rootCategoryPath != None):
            self.rootCategoryPath = rootCategoryPath
        else:
            self.rootCategoryPath = '-'
        """

        if root_category_path != None:
            category_chunks = root_category_path.split(" :: ")
            self.category = category_chunks[-1]
        else:
            self.category = "unknown"

    def _save_out(
        self,
        doc,
        reference_file_path,
        doc_name,
        doc_category,
        family_out_folder_path,
        session_id,
    ):
        """
        Saves a family to file if there is no match for it in the provided reference file. 
        The reference file is a FamilyBaseDataCombinedReport and is read into tuples using RevitFamilyBaseDataUtils.

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        :param referenceFilePath: Fully qualified file path to FamilyBaseDataCombinedReport file.
        :type referenceFilePath: str
        :param docName: The current family name.
        :type docName: str_
        :param docCategory: The family Revit category.
        :type docCategory: str
        :param familyOutFolderPath: The root directory path to which a family is saved. The script will create a sub directory\
            based on the revit batch processor session id and within that folder another sub directory based on the family Revit category:\
                familyOutFolderPath\\SessionId\\RevitCategory\\Myfamily.rfa
        :type familyOutFolderPath: str
        :param sessionId: The batchprocessor session Id (formatted so it can be used as a folder name)
        :type sessionId: str
        """

        # process reference file list and look for a match:
        # based on file name and category
        if fileIO.file_exist(reference_file_path):
            # read overall family base data from file
            (
                overall_family_base_root_data,
                overall_family_base_nested_data,
            ) = rFamBaseDataUtils.read_overall_family_data_list(reference_file_path)
            found_match = False
            for root_fam in overall_family_base_root_data:
                # check whether name and category are a match
                if root_fam.name == doc_name and root_fam.category == doc_category:
                    found_match = True
                    break
            # check if family needs saving out
            if found_match == False:
                # check session id folder exists
                if dirIO.create_target_directory(family_out_folder_path, session_id):
                    # check category folder exists
                    if dirIO.create_target_directory(
                        family_out_folder_path + "\\" + session_id, doc_category
                    ):
                        # save family out
                        rFile.save_as_family(
                            doc,
                            family_out_folder_path
                            + "\\"
                            + session_id
                            + "\\"
                            + doc_category,
                            doc_name,
                            [[doc_name, doc_name]],
                        )

    def process(self, doc, reference_file_path, family_out_directory_path, session_id):
        """
        Collects all base data from the document and stores it in the class property .data

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # get the family category name
        # famCatName = doc.OwnerFamily.FamilyCategory.Name

        # check if a reference file list was provided and if so if family needs to be saved out
        if (
            reference_file_path != None
            and family_out_directory_path != None
            and session_id != None
        ):
            self._save_out(
                doc,
                reference_file_path,
                self._strip_file_extension(doc.Title),
                self.category,
                family_out_directory_path,
                session_id,
            )

        # build data
        self.data.append(
            {
                IFamData.ROOT: self.root_path,
                IFamData.ROOT_CATEGORY: self.root_category_path,
                IFamData.FAMILY_NAME: self._strip_file_extension(doc.Title),
                IFamData.FAMILY_FILE_PATH: doc.PathName,  # this property will often be an empty string in nested families
                CATEGORY_NAME: self.category,
            }
        )

    def get_data(self):
        return self.data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A logger class
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Peter Smith
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import logging
import os
import sys

from duHast.Utilities.directory_io import create_target_directory
from duHast.Utilities.Objects.base import Base

class FilterFile(logging.Filter):
    """
    Filter to stop logs to file handler.

    To use change your log call to something like:
    my_logger.info("This is a console only output", extra={"block": "file"})
    """

    def filter(self, record):
        if "block" in record.__dict__.keys():
            if record.block == "file":
                return False
        return True


class FilterConsole(logging.Filter):
    """
    Filter to stop logs to console handler.

    To use change your log call to something like:
    my_logger.info("This is a file only output", extra={"block": "console"})
    """

    def filter(self, record):
        if "block" in record.__dict__.keys():
            if record.block == "console":
                return False
        return True


class LoggerObject(Base):
    def __init__(
        self,
        log_name="duHast",
        output_path=os.getenv("APPDATA"),
        log_level=(10, 30),
        **kwargs
    ):
        
        """
        Constructor for the LoggerObject class.

        :param log_name: The name of the log.
        :type log_name: str
        :param output_path: The path to the output directory.
        :type output_path: str
        :param log_level: The log levels for the file and console handlers.
        :type log_level: tuple
        """

        super(LoggerObject, self).__init__(**kwargs)

        # Get log naming and output path
        self.log_name = log_name
        self.output_path = output_path
        self.log_file_dir = os.path.join(self.output_path, self.log_name)
        create_target_directory(self.output_path, self.log_name)
        self.log_file_path = os.path.join(self.log_file_dir, self.log_name + ".log")
        # Get log levels
        self.log_level = log_level
        self.file_log_level, self.console_log_level = self.log_level
        # Establish handlers
        self.file_log_format = self.get_standard_formatter()
        self.file_handler = self.create_file_handler()
        self.console_log_format = self.get_presentation_formatter()
        self.console_handler = self.create_console_handler()
        # Create logger object
        new_logger = logging.getLogger(self.log_name)
        new_logger.propagate = 0
        while new_logger.handlers:
            new_logger.handlers.pop()
        new_logger.setLevel(logging.DEBUG)
        new_logger.addHandler(self.file_handler)
        new_logger.addHandler(self.console_handler)

        self.logger_object = new_logger
        logging.basicConfig(level=logging.DEBUG)

    def get_logger_obj(self):
        return self.logger_object

    def get_standard_formatter(self):
        """Standard output formatting for console and file handlers"""
        return logging.Formatter(
            "%(levelname)s | %(asctime)s.%(msecs)03d | %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def get_presentation_formatter(self):
        """Output formatting for presentation"""
        return logging.Formatter("%(message)s")

    def create_file_handler(self):
        """
        Create file handler to self.log_file_path location
        """
        # Set file output
        file_handler = logging.FileHandler(self.log_file_path)
        # Default file level is INFO
        file_handler.setLevel(self.file_log_level)
        # Add the filter to the file handler
        file_handler.addFilter(FilterFile())
        file_handler.setFormatter(self.file_log_format)
        file_handler.name = "file"
        return file_handler

    def create_console_handler(self):
        """
        Creates a console handler to output to stdout (console)
        displaying only the message as default
        """
        # Create a console handler
        console_handler = logging.StreamHandler(sys.stdout)
        # Default console level is WARNING
        console_handler.setLevel(self.console_log_level)
        # Add the filter to the console handler
        console_handler.addFilter(FilterConsole())
        console_handler.setFormatter(self.console_log_format)
        return console_handler

    def update_log_level(self, log_level):
        """
        Set the log levels for the file and console handlers
        """
        log_level_file, log_level_console = log_level
        self.file_handler.setLevel(log_level_file)
        self.console_handler.setLevel(log_level_console)
        return self.get_logger_obj()



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit links, CAD links and image links.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

def is_link_workset_visible(view, lnk_instance):
	"""
	Check if the workset of the link is visible in the view
	:param view: The view to check
	:type view: View
	:param lnk_instance: The Revit link instance to check the workset visibility of
	:type lnk_instance: RevitLinkInstance
	:return: True if the workset of the link is visible in the view, False otherwise
	:rtype: bool
	"""
	return view.IsWorksetVisible(lnk_instance.WorksetId)

def is_link_hidden(view, lnk_instance):
	"""
	Check if the link is hidden in the view
	:param view: The view to check
	:type view: View
	:param lnk_instance: The Revit link instance to check
	:type lnk_instance: RevitLinkInstance
	:return: True if the link is hidden in the view, False otherwise
	:rtype: bool
	"""
	return lnk_instance.IsHidden(view)

def is_link_visible(view_elem, lnk_inst):
	"""
	Check if the link is visible in the view. This is determined by:
	 - The link is not hidden
	 - The link workset is visible
	 
	:param view_elem: The view element to check
	:type view_elem: View
	:param lnk_inst: The Revit link instance to check
	:type lnk_inst: RevitLinkInstance
	:return: True if the link is visible in the view, False otherwise
	:rtype: bool
	"""
	link_is_not_hidden = is_link_hidden(view_elem, lnk_inst) == False
	link_workset_is_visible = is_link_workset_visible(view_elem, lnk_inst) == True
	return link_is_not_hidden and link_workset_is_visible


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Category data utility module containing functions to read category data from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reads family category data into  list of named tuples.

root_family:

- name 
- category 
- filePath 
- parent 
- child
- subcategories

nested_family:

- name 
- category
- filePath
- rootPath
- categoryPath
- hostFamily
- subcategories

reads change category directives into list of named tuples:

change_family_category:

- filePath
- newCategoryName

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from collections import namedtuple

from duHast.Utilities import (
    files_csv as fileCSV,
    files_get as fileGet,
    files_io as fileIO,
)

# tuples containing change family category data read from file
change_family_category = namedtuple(
    "change_family_category", "filePath newCategoryName"
)

# tuples containing change family subcategory data read from file
change_family_sub_category = namedtuple(
    "change_family_sub_category", "familyCategory oldSubCategoryName newSubCategoryName"
)

# tuples used to build category data
graphic_property_rgb = namedtuple("graphic_property_rgb", "red green blue")
graphic_property_line_weight = namedtuple(
    "graphic_property_line_weight", "cut projection"
)
graphic_property_material = namedtuple("graphic_property_material", "id name")
graphic_property_three_d_cut_projection = namedtuple(
    "graphicProperty", "threeD cut projection"
)
# container for category properties
sub_category_properties_container = namedtuple(
    "subCategoryProperties", "rgb lineWeight material graphic"
)
# the actual subcategory representing single row in report
sub_category = namedtuple(
    "sub_category",
    "parentCategoryName subCategoryName subCategoryId usageCounter usedBy subCategoryProperties ",
)

# a root family
root_family = namedtuple(
    "root_family", "name category filePath parent child subcategories"
)
# a nested family
nested_family = namedtuple(
    "nested_family",
    "name category filePath rootPath categoryPath hostFamily subcategories",
)

# row structure of family change category directive file
CATEGORY_CHANGE_DATA_LIST_INDEX_FAMILY_FILE_PATH = 0
CATEGORY_CHANGE_DATA_LIST_INDEX_NEW_FAMILY_CATEGORY = 1

# row structure of family change subcategory directive file
SUBCATEGORY_CHANGE_DATA_LIST_INDEX_FAMILY_CATEGORY = 0
SUBCATEGORY_CHANGE_DATA_LIST_INDEX_OLD_SUBCATEGORY_NAME = 1
SUBCATEGORY_CHANGE_DATA_LIST_INDEX_NEW_SUBCATEGORY_NAME = 2

# row structure of family category data file
CATEGORY_DATA_LIST_INDEX_ROOT_PATH = 0
CATEGORY_DATA_LIST_INDEX_ROOT_CATEGORY_PATH = 1
CATEGORY_DATA_LIST_INDEX_FAMILY_NAME = 2
CATEGORY_DATA_LIST_INDEX_FAMILY_FILE_PATH = 3
CATEGORY_DATA_LIST_INDEX_USAGE_COUNTER = 4
CATEGORY_DATA_LIST_INDEX_USED_BY = 5
CATEGORY_DATA_LIST_INDEX_CATEGORY_NAME = 6
CATEGORY_DATA_LIST_INDEX_SUBCATEGORY_NAME = 7
CATEGORY_DATA_LIST_INDEX_SUBCATEGORY_ID = 8
category_data_list_index_graphic_property_3_d = 9
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_CUT = 10
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_PROJECTION = 11
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_MATERIAL_NAME = 12
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_MATERIAL_ID = 13
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_LINE_WEIGHT_CUT = 14
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_LINE_WEIGHT_PROJECTION = 15
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_RGB_RED = 16
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_RGB_GREEN = 17
CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_RGB_BLUE = 18

# file name identifiers for family base data
FAMILY_CATEGORY_DATA_FILE_NAME_PREFIX = "FamilyCategories"
FAMILY_CATEGORY_DATA_FILE_EXTENSION = ".csv"

# file name identifiers for category change directives
CATEGORY_CHANGE_DIRECTIVE_FILE_NAME_PREFIX = "CategoryChangeDirective"
CATEGORY_CHANGE_DIRECTIVE_FILE_EXTENSION = ".csv"

# file name identifiers for subcategory change directives
SUBCATEGORY_CHANGE_DIRECTIVE_FILE_NAME_PREFIX = "SubCategoryChangeDirective"
SUBCATEGORY_CHANGE_DIRECTIVE_FILE_EXTENSION = ".csv"

# exceptions
EXCEPTION_NO_FAMILY_CHANGE_DIRECTIVE_DATA_FILES = (
    "Families change directive list files do not exist."
)
EXCEPTION_EMPTY_CHANGE_DIRECTIVE_DATA_FILES = (
    "Empty Families change directive data file(s)!"
)

EXCEPTION_NO_FAMILY_CATEGORY_DATA_FILES = (
    "Families category data list files do not exist."
)
EXCEPTION_EMPTY_FAMILY_CATEGORY_DATA_FILES = "Empty Families category data list file!"
EXCEPTION_NO_FAMILY_SUBCATEGORY_DATA_FILES = (
    "Families subcategory data list files do not exist."
)
EXCEPTION_EMPTY_FAMILY_SUBCATEGORY_DATA_FILES = (
    "Empty Families subcategory data list file!"
)


# -------------------------------- read category data set ----------------------------------------------------------------

# sample root and nested family tuple usage / set up
# set up subcategory properties
"""
dataRGB = graphic_property_rgb(120,120,120)
dataLineWeight = graphic_property_line_weight(3,6)
dataMaterial = graphic_property_material(-1,'')
dataGraphic = graphic_property_three_d_cut_projection(None,None,None)
# combine into a container
dataSubPropertiesContainer = sub_category_properties_container (dataRGB, dataLineWeight, dataMaterial, dataGraphic)
# set up the actual sub category ( single row in report )
dataSubCatSample = sub_category('Parent Cat Name', 'subCat name', 1234, dataSubPropertiesContainer)
# set up a root family tuple
dataRootFam = root_family('root family name', 'root family category', 'file path', [], [], [dataSubCatSample])
# set up a nested family tuple
dataNestedFam = nested_family('nested family name', 'nested family category', 'file path', 'root Path', 'category Path','host Family data', [dataSubCatSample])
"""
# end samples


def _create_root_family_from_data(data_row):
    """
    Sets up a root family tuple from data row past in.

    :param data_row: A row from the category report.
    :type data_row: [str]

    :return: a root_family tuple
    :rtype: named tuple :root_family
    """
    # need to check if this is a category belonging to the current family or a new family??
    fam = root_family(
        data_row[CATEGORY_DATA_LIST_INDEX_FAMILY_NAME],
        data_row[CATEGORY_DATA_LIST_INDEX_CATEGORY_NAME],
        data_row[CATEGORY_DATA_LIST_INDEX_FAMILY_FILE_PATH],
        [],  # set up an empty list for parent families
        [],  # set up an empty list for child families
        [],  # set up empty list for sub-categories
    )
    return fam


def _create_nested_family_from_data(data_row):
    """
    Sets up a nested family tuple from data row past in.

    :param data_row: A row from the category report.
    :type data_row: [str]

    :return: a nested family tuple
    :rtype: named tuple :nested_family
    """

    # found a child family
    fam = nested_family(
        data_row[CATEGORY_DATA_LIST_INDEX_FAMILY_NAME],
        data_row[CATEGORY_DATA_LIST_INDEX_CATEGORY_NAME],
        data_row[CATEGORY_DATA_LIST_INDEX_FAMILY_FILE_PATH],
        data_row[CATEGORY_DATA_LIST_INDEX_ROOT_PATH].split(
            " :: "
        ),  # split root path into list for ease of searching
        data_row[CATEGORY_DATA_LIST_INDEX_ROOT_CATEGORY_PATH].split(
            " :: "
        ),  # split category path into list for ease of searching
        [],  # set up an empty list for host families
        [],  # set up empty list for sub-categories
    )
    return fam


def _setup_family_from_data(data_row):
    """
    Creates a nested family or root family tuple from data row past in.

    :param data_row: A row from the category report.
    :type data_row: [str]

    :return: A nested or root family tuple.
    :rtype: named tuple
    """

    fam = None
    if "::" not in data_row[CATEGORY_DATA_LIST_INDEX_ROOT_PATH]:
        fam = _create_root_family_from_data(data_row)
    else:
        # found a child family
        fam = _create_nested_family_from_data(data_row)
    return fam


def _build_sub_category_properties_from_data(data_row):
    """
    Generates a subcategory tuple based on data row past in.

    :param data_row: A row from the category report.
    :type data_row: [str]

    :return: A sub_category tuple.
    :rtype: named tuple
    """

    # read category data first
    # get colour RGB values
    data_rgb = graphic_property_rgb(
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_RGB_RED],
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_RGB_GREEN],
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_RGB_BLUE],
    )
    # get line weight values
    data_line_weight = graphic_property_line_weight(
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_LINE_WEIGHT_CUT],
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_LINE_WEIGHT_PROJECTION],
    )
    # get material values
    data_material = graphic_property_material(
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_MATERIAL_ID],
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_MATERIAL_NAME],
    )
    # get graphic properties
    data_graphic = graphic_property_three_d_cut_projection(
        data_row[category_data_list_index_graphic_property_3_d],
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_CUT],
        data_row[CATEGORY_DATA_LIST_INDEX_GRAPHIC_PROPERTY_PROJECTION],
    )
    # put all of the above together
    data_sub_properties_container = sub_category_properties_container(
        data_rgb, data_line_weight, data_material, data_graphic
    )
    # set up the actual sub category ( single row in report )
    data_sub_cat_sample = sub_category(
        data_row[CATEGORY_DATA_LIST_INDEX_CATEGORY_NAME],
        data_row[CATEGORY_DATA_LIST_INDEX_SUBCATEGORY_NAME],
        data_row[CATEGORY_DATA_LIST_INDEX_SUBCATEGORY_ID],
        data_row[CATEGORY_DATA_LIST_INDEX_USAGE_COUNTER],
        data_row[CATEGORY_DATA_LIST_INDEX_USED_BY],
        data_sub_properties_container,
    )
    return data_sub_cat_sample


def _get_category_data_file_name(directory_path):
    """
    Gets the first family base data file in provided directory or any of it's sub directories.

    :param directory_path: Fully qualified directory path.
    :type directory_path: str
    :raises Exception: EXCEPTION_NO_FAMILY_BASE_DATA_FILES

    :return: Fully qualified file path to family base data file.
    :rtype: str
    """

    # get all base data files in folder
    files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        FAMILY_CATEGORY_DATA_FILE_NAME_PREFIX,
        "",
        FAMILY_CATEGORY_DATA_FILE_EXTENSION,
    )

    if len(files) > 0:
        return files[0]
    else:
        raise Exception(EXCEPTION_NO_FAMILY_CATEGORY_DATA_FILES)


def read_overall_family_data_list(file_path):
    """
    Reads list of families from family category data report file into named tuples.

    :param file_path: Fully qualified file path to family category data report file.
    :type file_path: str
    :raises Exception: "Families category data list files does not exist."
    :raises Exception: "Empty Families category data list file!"
    :return: Two lists: first list of named tuples contain family root data, second list contains family nested data.
    :rtype: [root_family], [nested_family]
    """

    rows = []
    if fileIO.file_exist(file_path):
        rows = fileCSV.read_csv_file(file_path)
    else:
        raise Exception(EXCEPTION_NO_FAMILY_CATEGORY_DATA_FILES)
    if len(rows) > 0:
        pass
    else:
        raise Exception(EXCEPTION_EMPTY_FAMILY_CATEGORY_DATA_FILES)

    return_value_root_family = []
    return_value_nested_family = []
    # pointer to the current family
    current_fam = None
    for i in range(1, len(rows)):
        # set up the actual sub category ( single row in report )
        data_sub_cat_sample = _build_sub_category_properties_from_data(rows[i])
        # get name and category as unique identifier
        fam_id = (
            rows[i][CATEGORY_DATA_LIST_INDEX_FAMILY_NAME]
            + rows[i][CATEGORY_DATA_LIST_INDEX_CATEGORY_NAME]
        )
        # check if this is the current family ...
        # this assumes family category data in report file is ordered by family!!!
        if current_fam == None:
            # and set up a new one:
            current_fam = _setup_family_from_data(rows[i])
            # append category data to new family
            current_fam.subcategories.append(data_sub_cat_sample)
        elif current_fam.name + current_fam.category == fam_id:
            # append category data to existing family
            current_fam.subcategories.append(data_sub_cat_sample)
        else:
            # if not:
            # append family to one of the list to be returned
            if isinstance(current_fam, root_family):
                return_value_root_family.append(current_fam)
            else:
                return_value_nested_family.append(current_fam)
            # and set up a new one:
            current_fam = _setup_family_from_data(rows[i])
            # append category data to new family
            current_fam.subcategories.append(data_sub_cat_sample)
    return return_value_root_family, return_value_nested_family


def read_overall_family_category_data_from_directory(directory_path):
    """
    Reads the first family category data file it finds in a folder.
    Note: This method calls ReadOverallFamilyDataList(filePath) which will raise exceptions if files are empty or dont exist in specified folder.

    :param directory_path: A fully qualified directory path containing family category data file(s)
    :type directory_path: _str

    :return: Two lists: first list of named tuples contain family root data, second list contains family nested data.
    :rtype: [root_family], [nested_family]
    """

    file_name = _get_category_data_file_name(directory_path)
    return read_overall_family_data_list(file_name)


# -------------------------------- read family category change directives ----------------------------------------------------------------


def read_overall_change_category_directives_list(file_paths):
    """
    Reads list of family change category directives from files into named tuples.

    :param file_path: List of fully qualified file path to family change category directive file.
    :type file_path: [str]
    :raises Exception: "Families change directive list files do not exist."
    :raises Exception: "Empty Families category data list file!"

    :return: List of named tuples contain family category change directive.
    :rtype: [change_family_category]
    """

    rows = []
    match_any_file = False
    for file_path in file_paths:
        if fileIO.file_exist(file_path):
            # set flag that we at least found one file
            match_any_file = True
            rows_file = fileCSV.read_csv_file(file_path)
            result = list(rows)
            result.extend(item for item in rows_file if item not in result)
            rows = result

    # check if any files found
    if match_any_file == False:
        raise Exception(EXCEPTION_NO_FAMILY_CHANGE_DIRECTIVE_DATA_FILES)

    # check if files contained any data
    if len(rows) > 0:
        # populate change directive tuples
        return_value_change_directives = []
        for row in rows:
            change_directive = change_family_category(
                row[CATEGORY_CHANGE_DATA_LIST_INDEX_FAMILY_FILE_PATH],
                row[CATEGORY_CHANGE_DATA_LIST_INDEX_NEW_FAMILY_CATEGORY],
            )
            return_value_change_directives.append(change_directive)
    else:
        raise Exception(EXCEPTION_EMPTY_CHANGE_DIRECTIVE_DATA_FILES)

    return return_value_change_directives


def _get_category_change_directive_file_names(directory_path):
    """
    Gets change category directive file in provided directory or any of it's sub directories.

    :param directory_path: Fully qualified directory path.
    :type directory_path: str
    :raises Exception: _EXCEPTION_NO_FAMILY_CHANGE_DIRECTIVE_DATA_FILES

    :return: List of fully qualified file path to family change category directive files.
    :rtype: [str]
    """

    # get all base data files in folder
    files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        CATEGORY_CHANGE_DIRECTIVE_FILE_NAME_PREFIX,
        "",
        CATEGORY_CHANGE_DIRECTIVE_FILE_EXTENSION,
    )
    if len(files) > 0:
        return files
    else:
        raise Exception(EXCEPTION_NO_FAMILY_CHANGE_DIRECTIVE_DATA_FILES)


def read_overall_family_category_change_directives_from_directory(directory_path):
    """
    Reads all category change directive file it finds in a folder.
    Note: This method calls ReadOverallFamilyDataList(filePath) which will raise exceptions if files are empty or dont exist in specified folder.

    :param directory_path: A fully qualified directory path containing family category change directive file(s)
    :type directory_path: _str

    :return: list of named tuples contain family category change directives.
    :rtype: [change_family_category]
    """

    file_names = _get_category_change_directive_file_names(directory_path)
    return read_overall_change_category_directives_list(file_names)


# -------------------------------- read family subcategory change directives ----------------------------------------------------------------


def read_overall_family_sub_category_change_directives_from_directory(directory_path):
    """
    Reads all subcategory change directive file it finds in a folder.
    Note: This method calls ReadOverallFamilyDataList(filePath) which will raise exceptions if files are empty or dont exist in specified folder.

    :param directory_path: A fully qualified directory path containing family category change directive file(s)
    :type directory_path: _str

    :return: list of named tuples contain family sub category change directives.
    :rtype: [change_family_sub_category]
    """

    file_names = _get_sub_category_change_directive_file_names(directory_path)
    return read_overall_change_sub_category_directives_list(file_names)


def _get_sub_category_change_directive_file_names(directory_path):
    """
    Gets change subcategory directive file in provided directory or any of it's sub directories.

    :param directory_path: Fully qualified directory path.
    :type directory_path: str
    :raises Exception: _EXCEPTION_NO_FAMILY_CHANGE_DIRECTIVE_DATA_FILES

    :return: List of fully qualified file path to family change sub category directive files.
    :rtype: [str]
    """

    # get all base data files in folder
    files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        SUBCATEGORY_CHANGE_DIRECTIVE_FILE_NAME_PREFIX,
        "",
        SUBCATEGORY_CHANGE_DIRECTIVE_FILE_EXTENSION,
    )
    if len(files) > 0:
        return files
    else:
        raise Exception(EXCEPTION_NO_FAMILY_CHANGE_DIRECTIVE_DATA_FILES)


def read_overall_change_sub_category_directives_list(file_paths):
    """
    Reads list of family change subcategory directives from files into named tuples.

    :param file_path: List of fully qualified file path to family change category directive file.
    :type file_path: [str]
    :raises Exception: _EXCEPTION_NO_FAMILY_SUBCATEGORY_DATA_FILES
    :raises Exception: _EXCEPTION_EMPTY_FAMILY_SUBCATEGORY_DATA_FILES

    :return: List of named tuples containing family subcategory change directive.
    :rtype: [change_family_sub_category]
    """

    rows = []
    match_any_file = False
    for file_path in file_paths:
        if fileIO.file_exist(file_path):
            # set flag that we at least found one file
            match_any_file = True
            rows_file = fileCSV.read_csv_file(file_path)
            result = list(rows)
            result.extend(item for item in rows_file if item not in result)
            rows = result

    # check if any files found
    if match_any_file == False:
        raise Exception(EXCEPTION_NO_FAMILY_SUBCATEGORY_DATA_FILES)

    # check if files contained any data
    if len(rows) > 0:
        # populate change directive tuples
        return_value_change_directives = []
        for row in rows:
            change_directive = change_family_sub_category(
                row[SUBCATEGORY_CHANGE_DATA_LIST_INDEX_FAMILY_CATEGORY],
                row[SUBCATEGORY_CHANGE_DATA_LIST_INDEX_OLD_SUBCATEGORY_NAME],
                row[SUBCATEGORY_CHANGE_DATA_LIST_INDEX_NEW_SUBCATEGORY_NAME],
            )
            return_value_change_directives.append(change_directive)
    else:
        raise Exception(EXCEPTION_EMPTY_FAMILY_SUBCATEGORY_DATA_FILES)

    return return_value_change_directives


def get_families_requiring_sub_category_change(
    root_families, sub_cat_change_directives
):
    """
    Returns a list of file path of root families containing subcategories requiring a rename.

    Note: list of file path returned is unique: i.e.  if a family has multiple matches for rename subcategory directives it will still only appear once in the list.

    :param root_families: List of tuples of root families.
    :type root_families: [root_family]
    :param sub_cat_change_directives: List of subcategory change directives.
    :type sub_cat_change_directives: [change_family_sub_category]

    :return: List of family file path.
    :rtype: [str]
    """

    root_families_needing_change = []
    # check each root family
    for root_fam in root_families:
        # against each subcategory change directive
        for change_d in sub_cat_change_directives:
            # check if match family category
            if root_fam.category == change_d.familyCategory:
                # loop over subcategories in family and check if any of them needs renaming
                for sub_cat_root in root_fam.subcategories:
                    if sub_cat_root.subCategoryName == change_d.oldSubCategoryName:
                        # found a match
                        if root_fam.filePath not in root_families_needing_change:
                            root_families_needing_change.append(root_fam.filePath)
                        break
    return root_families_needing_change



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit API utility functions to get parameter values.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import System
from System import Type
import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import everything from Autodesk Revit DataBase namespace (Revit API)
from Autodesk.Revit.DB import ElementId  # , ParameterType, StorageType
import Autodesk.Revit.DB as rdb

# utilities
from duHast.Utilities import utility as util
from duHast.Utilities import unit_conversion as unitConversion

# type checker
# from typing import List, Callable


def has_property(obj, property_name):
    """
    Reflection to work out whether a property on a .net object exists (in case it has been removed)

    :param obj: A .net object
    :type obj: var
    :param property_name: The property name to check for
    :type property_name: str
    :return: None if the property does not exist
    :rtype: _type_
    """

    obj_type = obj.GetType()
    prop_info = obj_type.GetProperty(property_name)
    return prop_info is not None


# ----------------------------------------parameters-----------------------------------------------


def check_parameter_value(
    para,
    para_condition,  # type: Callable[[str,str],bool]
    condition_value,
):
    # type: (...) -> bool
    """
    Checks a parameter value based on passed in condition function.

    This extracts the value of the past in parameter and compares it against a past in value using
    the also past in compare function.
    Note that values will be past into compare function as ASCII encoded.

    :param para: Parameter of which the value is to be checked.
    :type para: Autodesk.Revit.DB.Parameter
    :param para_condition:
        Function taking 2 arguments:
        First argument is the value to be checked against
        Second argument is the actual parameter value
        Needs to return a bool!
        Both arguments will be ASCII encoded at passing in.
    :type para_condition: function
    :param condition_value: The value to be checked for
    :type condition_value: var
    :raise: Any exception will need to be managed by the function caller.

    :return:
        True if condition value is evaluated to be True by past in function paraCondition.
        Will return False if compare function returns None or a False.
    :rtype: bool

    """
    # set default return value
    is_match = False
    parameter_value = get_parameter_value(para)
    # evaluate parameter value with past in value using past in function
    compare_outcome = para_condition(
        util.encode_ascii(condition_value), util.encode_ascii(parameter_value)
    )
    # check the return value for a bool (True) only. Everything else will return False
    if compare_outcome == True:
        is_match = True
    return is_match


# ----------------------------------------parameters value getter over loads-----------------------------------------------


def getter_none(para):
    """
    Used for parameters where the storage type is None

    :param para: _description_
    :type para: _type_
    """

    return "Invalid storage type: (NONE)"


def getter_double_or_int_as_string(para):
    """
    Returns a parameter value of type double or integer as string.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: String representation of double or integer value. If value is empty it will return None
    :rtype: str or None
    """

    parameter_value = "None"
    if para.AsValueString() != None and para.AsValueString() != "":
        parameter_value = para.AsValueString()
    return parameter_value


def getter_double_as_double(para):
    """
    Returns a parameter value of type double as a double.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: Double value. If value is empty it will return None
    :rtype: Double or None
    """

    parameter_value = None
    if para.AsValueString() != None and para.AsValueString() != "":
        parameter_value = para.AsDouble()
    return parameter_value


def getter_double_as_double_converted_to_metric(para):
    """
    Returns a parameter value of type double to metric if required.
    Revit uses feet internally for any length value.
    Revit uses square feet for areas

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: Double value. If value is empty it will return None
    :rtype: Double or None
    """

    parameter_value = None

    # in revit 2023 Definition.ParameterType is obsolete...check for it:
    if has_property(obj=para.Definition, property_name="ParameterType"):
        if para.AsValueString() != None and para.AsValueString() != "":
            if para.Definition.ParameterType == rdb.ParameterType.Length:
                parameter_value = unitConversion.convert_imperial_feet_to_metric_mm(
                    para.AsDouble()
                )
            elif para.Definition.ParameterType == rdb.ParameterType.Area:
                parameter_value = (
                    unitConversion.convert_imperial_square_feet_to_metric_square_metre(
                        para.AsDouble()
                    )
                )
            elif para.Definition.ParameterType == rdb.ParameterType.Volume:
                parameter_value = (
                    unitConversion.convert_imperial_cubic_feet_to_metric_cubic_metre(
                        para.AsDouble()
                    )
                )
            else:
                parameter_value = para.AsDouble()
    else:
        # assume this is revit 2023 onwards
        data_type = para.Definition.GetDataType()
        if para.AsValueString() != None and para.AsValueString() != "":
            if data_type == rdb.SpecTypeId.Length:
                parameter_value = unitConversion.convert_imperial_feet_to_metric_mm(
                    para.AsDouble()
                )
            elif data_type == rdb.SpecTypeId.Area:
                parameter_value = (
                    unitConversion.convert_imperial_square_feet_to_metric_square_metre(
                        para.AsDouble()
                    )
                )
            elif data_type == rdb.SpecTypeId.Volume:
                parameter_value = (
                    unitConversion.convert_imperial_cubic_feet_to_metric_cubic_metre(
                        para.AsDouble()
                    )
                )
            else:
                parameter_value = para.AsDouble()

    return parameter_value


def getter_int_as_int(para):
    """
    Returns a parameter value of type integer as a integer.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: Integer value. If value is empty it will return None
    :rtype: Integer or None
    """

    parameter_value = None
    if para.AsValueString() != None and para.AsValueString() != "":
        parameter_value = para.AsInteger()
    return parameter_value


def getter_string_as_UTF8_string(para):
    """
    Returns a parameter value of type string as a utf-8 formatted string.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: String value. If value is empty it will return None
    :rtype: String or None
    """

    parameter_value = "None"
    if para.StorageType == rdb.StorageType.String:
        if para.AsString() != None and para.AsString() != "":
            parameter_value = para.AsString().encode("utf-8")
    return parameter_value


def getter_string_as_string(para):
    """
    Returns a parameter value of type string as a string.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: String value. If value is empty it will return None
    :rtype: String or None
    """

    parameter_value = "None"
    if para.StorageType == rdb.StorageType.String:
        if para.AsString() != None and para.AsString() != "":
            parameter_value = para.AsString()
    return parameter_value


def getter_element_id_as_string(para):
    """
    Returns a parameter value of type element id as a string.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: String value. If value is empty it will return None
    :rtype: String or None
    """

    parameter_value = "None"
    if para.StorageType == rdb.StorageType.ElementId:
        if para.AsElementId() != None:
            parameter_value = str(para.AsElementId())
    return parameter_value


def getter_element_id_as_element_id(para):
    """
    Returns a parameter value of type element id as a element id.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: Element id value. If value is empty it will return None
    :rtype: Element id or None
    """

    parameter_value = None
    if para.StorageType == rdb.StorageType.ElementId:
        if para.AsElementId() != None:
            parameter_value = para.AsElementId()
    return parameter_value


def getter_element_id_as_element_int(para):
    """
    Returns a parameter value of type element id as an integer.

    :param para: The parameter.
    :type para: Autodesk.Revit.DB.Parameter

    :return: Integer value. If value is empty it will return None
    :rtype: Integer or None
    """

    parameter_value = None
    if para.StorageType == rdb.StorageType.ElementId:
        if para.AsElementId() != None:
            parameter_value = para.AsElementId().IntegerValue
    return parameter_value


# ----------------------------------------parameters value getters -----------------------------------------------


def get_parameter_value_with_over_load(para, parameter_value_getters):
    """
    Returns a parameter value in format depending on storage type.

    Storage type can be:

    - Double
    - Integer
    - String
    - ElementId

    Will throw an exception if a storage type is not covered by parameter value getter functions.

    :param para: The Parameter.
    :type para: Autodesk.Revit.DB.Parameter
    :param parameter_value_getters: Dictionary containing the functions returning the parameter value depending on parameter storage type
    :type parameter_value_getters: {Autodesk.Revit.DB.StorageType: func()}
    :return: The parameter value or if empty: None.
    :rtype: Depends on value getters functions
    """

    # set return value default ( default value should never be used...!)
    parameter_value = None
    try:
        # extract parameter value depending on its storage type
        if para.StorageType == rdb.StorageType.Double:
            if rdb.StorageType.Double in parameter_value_getters:
                parameter_value = parameter_value_getters[rdb.StorageType.Double](para)
            else:
                raise ValueError(
                    "No parameter value getter for storage type Double provided"
                )
        elif para.StorageType == rdb.StorageType.Integer:
            if rdb.StorageType.Integer in parameter_value_getters:
                parameter_value = parameter_value_getters[rdb.StorageType.Integer](para)
            else:
                raise ValueError(
                    "No parameter value getter for storage type Integer provided"
                )
        elif para.StorageType == rdb.StorageType.String:
            if rdb.StorageType.String in parameter_value_getters:
                parameter_value = parameter_value_getters[rdb.StorageType.String](para)
            else:
                raise ValueError(
                    "No parameter value getter for storage type String provided"
                )
        elif para.StorageType == rdb.StorageType.ElementId:
            if rdb.StorageType.ElementId in parameter_value_getters:
                parameter_value = parameter_value_getters[rdb.StorageType.ElementId](
                    para
                )
            else:
                raise ValueError(
                    "No parameter value getter for storage type Element Id provided"
                )
        else:
            # this should be invalid storage type only
            parameter_value = parameter_value_getters[str(None)](para)

    except Exception as e:
        parameter_value = "Exception: {}".format(e)
    return parameter_value


def get_parameter_value(para):
    # type: (...) -> str
    """
    Returns a parameter value as string independent of its storage type.

    :param para: Parameter of which the value is to be returned.
    :type para: Autodesk.Revit.DB.Parameter
    :raise: If an exception occurs the exception message will be returned as the parameter value prefixed with 'Exception: '

    :return:
        Default value is 'no Value' if parameter value is empty.
        Otherwise the actual parameter value.
        Will return 'Exception: ' + exception message if an exception occurred.
    :rtype: str
    """

    # set return value default
    parameter_value = "no Value"
    try:
        value_getter = {
            rdb.StorageType.Double: getter_double_or_int_as_string,
            rdb.StorageType.Integer: getter_double_or_int_as_string,
            rdb.StorageType.String: getter_string_as_string,
            rdb.StorageType.ElementId: getter_element_id_as_string,
            str(None): getter_none,
        }

        parameter_value = get_parameter_value_with_over_load(para, value_getter)

    except Exception as e:
        parameter_value = "Exception: " + str(e)

    return parameter_value


def get_parameter_value_utf8_string(para):
    # type: (...) -> str
    """
    Returns the parameter value as utf-8 string independent of its storage type.

    :param para: Parameter of which the value is to be returned.
    :type para: Autodesk.Revit.DB.Parameter
    :raise: If an exception occurs the exception message will be returned as the parameter value prefixed with 'Exception: '

    :return:
        Default value is 'no Value' if parameter value is empty.
        Otherwise the actual parameter value.
        Will return 'Exception: ' + exception message if an exception occurred.
    :rtype: str
    """

    # set return value default
    parameter_value = "no Value"

    value_getter = {
        rdb.StorageType.Double: getter_double_or_int_as_string,  # no specific utf encoding required
        rdb.StorageType.Integer: getter_double_or_int_as_string,  # no specific utf encoding required
        rdb.StorageType.String: getter_string_as_UTF8_string,
        rdb.StorageType.ElementId: getter_element_id_as_string,  # no specific utf encoding required
        str(None): getter_none,
    }

    parameter_value = get_parameter_value_with_over_load(para, value_getter)

    return parameter_value


def get_parameter_value_as_integer(para):
    # type: (...) -> int
    """
    Returns the parameter value as integer only if the storage type is integer. Otherwise -1 will be returned.

    :param para: Parameter of which the value is to be returned.
    :type para: Autodesk.Revit.DB.Parameter
    :raise: Any exception will need to be managed by the function caller.

    :return:
        Default value is -1 if parameter value is empty or storage type is not integer.
        Otherwise the actual parameter value.
    :rtype: int
    """

    # set return value default
    parameter_value = -1

    value_getter = {rdb.StorageType.Integer: getter_int_as_int}

    # extract parameter value depending on whether its storage type is integer, otherwise default value
    if para.StorageType == rdb.StorageType.Integer:
        parameter_value = get_parameter_value_with_over_load(para, value_getter)
    return parameter_value


def get_parameter_value_as_element_id(para):
    """
    Returns the parameter value as element Id only if the storage type is ElementId. Otherwise InvalidElementId (-1) will be returned.

    :param para: Parameter of which the value is to be returned.
    :type para: Autodesk.Revit.DB.Parameter
    :raise: Any exception will need to be managed by the function caller.

    :return:
        Default value is -1 if parameter value is empty or storage type is not integer.
        Otherwise the actual parameter value.
    :rtype: int
    """

    # set return value default
    parameter_value = ElementId.InvalidElementId

    value_getter = {rdb.StorageType.ElementId: getter_element_id_as_element_id}

    # check if element id...otherwise return the default value
    if para.StorageType == rdb.StorageType.ElementId:
        parameter_value = get_parameter_value_with_over_load(para, value_getter)
    return parameter_value


def get_all_parameters_and_values_wit_custom_getters(element, parameter_value_getters):
    """
    Returns all parameters and their values as using custom value getter associated with provided element in form of a dictionary.

    :param element: The element
    :type element: var
    :param parameter_value_getters: Dictionary containing the functions returning the parameter value depending on parameter storage type
    :type parameter_value_getters: {Autodesk.Revit.DB.StorageType: func()}

    :return: Dictionary where key is the parameter name, and the value is the parameter value.
    :rtype: {str:var}
    """
    return_value = {}
    paras = element.GetOrderedParameters()
    for p in paras:
        p_value = get_parameter_value_with_over_load(p, parameter_value_getters)
        return_value[p.Definition.Name] = p_value
    return return_value


def get_built_in_parameter_value(
    element,
    built_in_parameter_def,
    parameter_value_getter=get_parameter_value_utf8_string,
):
    """
    Returns the built-in parameter value. Return value type depends on past in value getter function. Default is UTF-8 encoded string.

    :param element: Element to which the built-in parameter belongs.
    :type element: Autodesk.Revit.DB.Element
    :param built_in_parameter_def: The parameters built-in definition of which the value is to be returned.
    :type built_in_parameter_def: Autodesk.Revit.DB.Definition
    :param parameter_value_getter:
        The function which takes the parameter as an argument and returns it's value.
    :type parameter_value_getter: function
    :raise: As per value getter method.

    :return:
        Default value is None if parameter does not exist on element.
        Otherwise the actual parameter value as per value getter method.
    :rtype: var
    """

    # set return value default
    parameter_value = None
    para = element.get_Parameter(built_in_parameter_def)
    if para != None:
        parameter_value = parameter_value_getter(para)

    return parameter_value


def get_parameter_value_by_name(
    element,
    parameter_name,  # type: str
    parameter_value_getter=get_parameter_value_utf8_string,
):
    """
    Returns the parameter value by parameter name.

    Return value type depends on past in value getter function. Default is UTF-8 encoded string.

    :param element: Element to which the built-in parameter belongs.
    :type element: Autodesk.Revit.DB.Element
    :param parameter_name: The parameters name of which the value is to be returned.
    :type parameter_name: str
    :param parameter_value_getter:
        The function which takes the parameter as an argument and returns it's value.
    :type parameter_value_getter: function
    :raise: As per value getter method.

    :return:
        Default value is None if parameter does not exist on element.
        Otherwise the actual parameter value as per value getter method.
    :rtype: var
    """

    # set return value default
    parameter_value = None
    para = element.LookupParameter(parameter_name)
    if para != None:
        parameter_value = parameter_value_getter(para)
    return parameter_value


def param_is_empty_or_null(param):
    """
    Checks if a parameter is empty or null
    :param param: The parameter to check
    :type param: Parameter
    :return: True if the parameter is empty or null, False if it is not
    :rtype: bool
    """

    if param == None:
        return True
    else:
        if param.AsString() == "" or param.AsString() == None:
            return True
        else:
            return False



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains utility functions - hash tables from view settings.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# defines fixed hash values for ease of identification!
# no override is present
NO_OVERRIDE = 0
# category is switched of or filtered elements are switched off
SWITCHED_OFF = 1
# filter is not enabled
FILTER_NOT_ENABLED = 2
# filter or category does not exist in view or model
DOES_NOT_EXIST = -1


def _get_hash_headers(views_settings):
    """
    Headers are the Revit view template names

    :param views_settings: An instance of view graphic settings class storing Revit view template data.
    :type views_settings: :class:`.ViewGraphicsSettings`
    :return: List of headers representing all view template names in a model.
    :rtype: [str]
    """
    headers = []
    for setting in views_settings:
        headers.append(setting.view_name)
    return headers


def _get_hash_rows_categories(views_settings):
    """
    Rows are the Revit category names

    :param views_settings: An instance of view graphic settings class storing Revit view template data.
    :type views_settings: :class:`.ViewGraphicsSettings`
    :return: List of row headers representing all object categories in a model.
    :rtype: [str]
    """

    rows = []
    for override in views_settings[0].override_by_category:
        rows.append(override.main_category_name + " :: " + override.sub_category_name)
    return rows


def _get_hash_rows_filters(views_settings):
    """
    Rows are all unique filter names

    :param views_settings: An instance of view graphic settings class storing Revit view template data.
    :type views_settings: :class:`.ViewGraphicsSettings`
    """

    rows = []
    for setting in views_settings:
        for override in setting.override_by_filter:
            if override.filter_name not in rows:
                rows.append(override.filter_name)
    return rows


def _get_hash_for_category_overrides(headers, row_headers, views_settings):
    """
    Returns a list with a hash values for each Revit category override in a given view template.

    :param headers: List of headers representing all view template names in a model.
    :type headers: [str]
    :param row_headers: List of row headers representing all object categories in a model.
    :type row_headers: [str]
    :param views_settings: An instance of view graphic settings class storing Revit view template data.
    :type views_settings: :class:`.ViewGraphicsSettings`
    :raises ValueError: _description_
    :return: List of hash values.
    :rtype: [int]
    """

    # loop over header and get the respective view override setting
    # loop over rows and get the hash of the row value
    # return the hash table
    table_hash = {}
    for header_view_name in headers:
        # find the view (template)
        matching_template = next(
            (
                instance
                for instance in views_settings
                if instance.view_name == header_view_name
            ),
            None,
        )
        if matching_template == None:
            raise ValueError(
                "Impossible!!! No match found for view: [{}]".format(header_view_name)
            )

        # get category hashes
        for row in row_headers:
            # split to get main and sub category name
            name_parts = row.split(" :: ", 1)
            # Find a matching instance based on the two properties using a lambda expression and filter function
            matching_category = next(
                (
                    instance
                    for instance in matching_template.override_by_category
                    if instance.main_category_name == name_parts[0]
                    and instance.sub_category_name == name_parts[1]
                ),
                None,
            )

            # set a default value
            hash_value = DOES_NOT_EXIST
            # get the actual hash value if a match was found
            if matching_category:
                # check if an override is present and whether that the category is visible )
                # If that is the case return a hash of all properties

                if matching_category.is_visible == False:
                    # category is not visible
                    hash_value = SWITCHED_OFF
                elif (
                    matching_category.are_overrides_present == False
                    and matching_category.is_visible == True
                ):
                    # category is visible but no override is applied
                    hash_value = NO_OVERRIDE
                else:
                    # category is visible and an override is applied
                    hash_value = hash(matching_category)

            # add to table
            if row in table_hash:
                table_hash[row].append(hash_value)
            else:
                table_hash[row] = [hash_value]

    # convert into a simplified table where:
    # column headers are view names
    # row headers are category names
    simple_table = []
    for key, value in table_hash.items():
        simple_table.append(value)

    return simple_table


def _get_hash_for_filter_overrides(headers, row_headers, views_settings):
    """
    Returns a list with a hash values for each view filter override in a given view template.

    :param headers: List of headers representing all view template names in a model.
    :type headers: [str]
    :param row_headers: List of row headers representing each filter applied in a Revit view template.
    :type row_headers: [str]]
    :param views_settings: An instance of view graphic settings class storing Revit view template data.
    :type views_settings: :class:`.ViewGraphicsSettings`
    :raises ValueError: _description_
    :return: List of hash values.
    :rtype: [int]
    """

    # loop over header and get the respective view override setting
    # loop over rows and get the hash of the row value
    # return the hash table
    table_hash = {}
    for header_view_name in headers:
        # find the view (template)
        matching_template = next(
            (
                instance
                for instance in views_settings
                if instance.view_name == header_view_name
            ),
            None,
        )
        if matching_template == None:
            raise ValueError(
                "Impossible!!! No match found for view [{}]".format(header_view_name)
            )

        # get category hashes
        for row in row_headers:
            # Find a matching instance based on the two properties using a lambda expression and filter function
            matching_filter = next(
                (
                    instance
                    for instance in matching_template.override_by_filter
                    if instance.filter_name == row
                ),
                None,
            )

            # set a default value
            hash_value = DOES_NOT_EXIST
            # get the actual hash value if a match was found
            if matching_filter:
                # check if an override is present and whether that the elements filtered are visible )
                # If that is the case return a hash of all properties

                if matching_filter.is_visible == False:
                    # elements filtered are not visible
                    hash_value = SWITCHED_OFF
                elif (
                    matching_filter.are_overrides_present == False
                    and matching_filter.is_visible == True
                ):
                    # elements filtered are visible but no override is present
                    hash_value = NO_OVERRIDE
                elif (
                    matching_filter.are_overrides_present == False
                    and matching_filter.is_visible == True
                    and matching_filter.is_enabled == False
                ):
                    hash_value = FILTER_NOT_ENABLED
                else:
                    # elements filtered are visible and an override is present and the filter is enabled
                    hash_value = hash(matching_filter)

            # add to table
            if row in table_hash:
                table_hash[row].append(hash_value)
            else:
                table_hash[row] = [hash_value]

    # convert into a simplified table where:
    # column headers are view names
    # row headers are category names
    simple_table = []
    for key, value in table_hash.items():
        simple_table.append(value)

    return simple_table



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit sub-category property set functions .
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.LinePattern import line_patterns as rPat
from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res
from duHast.Revit.Categories.Utility.category_property_names import (
    PROPERTY_LINE_COLOUR_BLUE_NAME,
    PROPERTY_LINE_COLOUR_GREEN_NAME,
    PROPERTY_LINE_COLOUR_RED_NAME,
    PROPERTY_LINE_WEIGHT_CUT_NAME,
    PROPERTY_LINE_WEIGHT_PROJECTION_NAME,
    PROPERTY_MATERIAL_ID,
)
from duHast.Revit.Categories.Utility.category_properties_get_utils import (
    get_saved_category_property_by_name,
)


def set_category_material(doc, cat, material_id):
    """
    Updates material property of a given category.
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :param material_id: The new material element id.
    :type material_id: Autodesk.Revit.DB.ElementId
    :return: True if material property was updated successfully, otherwise False.
    :rtype: bool
    """

    flag = True
    try:
        mat = doc.GetElement(material_id)

        def action():
            action_return_value = res.Result()
            try:
                cat.Material = mat
                action_return_value.update_sep(
                    True, "Successfully set material value of subcategory"
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set material value of subcategory with exception: {}".format(
                        e
                    ),
                )
            return action_return_value

        transaction = rdb.Transaction(
            doc, "Updating subcategory material: " + str(rdb.Element.Name.GetValue(mat))
        )
        update_mat = rTran.in_transaction(transaction, action)
        flag = update_mat.status
    except Exception as e:
        flag = False
    return flag


def set_category_line_pattern(doc, cat, line_pattern_id, ignore_missing_cut_style):
    """
    Updates line pattern property of a given category.
    Note: in cases where the 'cut' property does not exist on a sub category this will return false even though the 'projection' property will most
    likely have been updated without a problem...
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :param line_pattern_id: The newline pattern element id.
    :type line_pattern_id: Autodesk.Revit.DB.ElementId
    :param ignore_missing_cut_style: If true will not flag an exception if applying styles fails on missing cut style.
    :type ignore_missing_cut_style: bool
    :return: True if line pattern property was updated successfully, otherwise False.
    :rtype: bool
    """

    flag = True
    try:

        def action():
            action_return_value = res.Result()
            try:
                cat.SetLinePatternId(line_pattern_id, rdb.GraphicsStyleType.Cut)
                action_return_value.update_sep(
                    True, "Successfully set cut line pattern of subcategory"
                )
            except Exception as e:
                if ignore_missing_cut_style:
                    action_return_value.update_sep(
                        True,
                        "Failed to set cut line pattern of subcategory with exception: {}. Exception ignored!".format(
                            e
                        ),
                    )
                else:
                    action_return_value.update_sep(
                        False,
                        "Failed to set cut line pattern of subcategory with exception: {}".format(
                            e
                        ),
                    )
            try:
                cat.SetLinePatternId(line_pattern_id, rdb.GraphicsStyleType.Projection)
                action_return_value.update_sep(
                    True, "Successfully set projection line pattern of subcategory"
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set projection line pattern of subcategory with exception: {}".format(
                        e
                    ),
                )
            return action_return_value

        transaction = rdb.Transaction(doc, "Updating subcategory line pattern")
        update_line_pattern = rTran.in_transaction(transaction, action)
        flag = update_line_pattern.status
    except Exception as e:
        flag = False
    return flag


def set_category_line_weights(
    doc, cat, line_thick_ness_cut, line_thickness_projection, ignore_missing_cut_style
):
    """
    Updates line weight properties of a given category.
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :param line_thick_ness_cut: The cut line weight.
    :type line_thick_ness_cut: int
    :param line_thickness_projection: The projection line weight.
    :type line_thickness_projection: int
    :param ignore_missing_cut_style: If true will not flag an exception if applying styles fails on missing cut style.
    :type ignore_missing_cut_style: bool
    :return: True if line weight property was updated successfully, otherwise False.
    :rtype: bool
    """

    flag = True
    try:

        def action():
            action_return_value = res.Result()
            try:
                cat.SetLineWeight(line_thick_ness_cut, rdb.GraphicsStyleType.Cut)
                action_return_value.update_sep(
                    True, "Successfully set cut line weight of subcategory"
                )
            except Exception as e:
                if ignore_missing_cut_style:
                    action_return_value.update_sep(
                        True,
                        "Failed to set cut line weight of subcategory with exception: {}. Exception ignored!".format(
                            e
                        ),
                    )
                else:
                    action_return_value.update_sep(
                        False,
                        "Failed to set cut line weight of subcategory with exception: {}".format(
                            e
                        ),
                    )
            try:
                cat.SetLineWeight(
                    line_thickness_projection, rdb.GraphicsStyleType.Projection
                )
                action_return_value.update_sep(
                    True, "Successfully set projection line weight of subcategory"
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set projection line weight of subcategory with exception: {}".format(
                        e
                    ),
                )
            return action_return_value

        transaction = rdb.Transaction(doc, "Updating subcategory line weights")
        update_line_weights = rTran.in_transaction(transaction, action)
        flag = update_line_weights.status
    except Exception as e:
        flag = False
    return flag


def set_category_colour(doc, cat, red, green, blue):
    """
    Updates colour properties of a given category.
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :param red: The colour red channel.
    :type red: byte
    :param green: The colour green channel.
    :type green: byte
    :param blue: The colour blue channel.
    :type blue: byte
    :return: True if colour property was updated successfully, otherwise False.
    :rtype: bool
    """

    flag = True
    try:

        def action():
            action_return_value = res.Result()
            try:
                new_colour = rdb.Color(red, green, blue)
                cat.LineColor = new_colour
                action_return_value.update_sep(
                    True, "Successfully set colour value of subcategory"
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set colour value of subcategory with exception: {}".format(
                        e
                    ),
                )
            return action_return_value

        transaction = rdb.Transaction(doc, "Updating subcategory colour")
        update_colour = rTran.in_transaction(transaction, action)
        flag = update_colour.status
    except Exception as e:
        flag = False
    return flag


def set_category_properties(doc, cat, properties, ignore_missing_cut_style):
    """
    Updates varies property values of a given category.
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :param properties: List of property values to be applied to category.
    :type properties: list of dictionaries in format as per GetCategoryProperties(cat) method.
    :param ignore_missing_cut_style: If true will not flag an exception if applying styles fails on missing cut style.
    :type ignore_missing_cut_style: bool
    :return: True if all properties where updated successfully, otherwise False.
    :rtype: bool
    """

    # material
    mat_id = get_saved_category_property_by_name(properties, [PROPERTY_MATERIAL_ID])
    flag_mat = set_category_material(doc, cat, mat_id[0])

    # line pattern
    line_pattern_id = get_saved_category_property_by_name(
        properties, [rPat.PROPERTY_PATTERN_ID]
    )
    flag_pattern = set_category_line_pattern(
        doc, cat, line_pattern_id[0], ignore_missing_cut_style
    )

    # line weights
    line_weights = get_saved_category_property_by_name(
        properties,
        [PROPERTY_LINE_WEIGHT_CUT_NAME, PROPERTY_LINE_WEIGHT_PROJECTION_NAME],
    )
    flag_line_weights = set_category_line_weights(
        doc, cat, line_weights[0], line_weights[1], ignore_missing_cut_style
    )

    # category colour
    colour_rgb = get_saved_category_property_by_name(
        properties,
        [
            PROPERTY_LINE_COLOUR_RED_NAME,
            PROPERTY_LINE_COLOUR_GREEN_NAME,
            PROPERTY_LINE_COLOUR_BLUE_NAME,
        ],
    )
    flag_colours = set_category_colour(
        doc, cat, colour_rgb[0], colour_rgb[1], colour_rgb[2]
    )

    return flag_mat & flag_pattern & flag_line_weights & flag_colours



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit door export to DATA class functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import BoundingBoxXYZ, BuiltInParameter, Element, Options, StorageType

from duHast.Revit.Common import (
    design_set_options as rDesignO,
    parameter_get_utils as rParaGet,
    phases as rPhase,
)

# from duHast.APISamples.Ceilings.Geometry import Geometry
from duHast.Revit.Doors.doors import get_door_instances
from duHast.Data.Objects import data_door as dDoor
from duHast.Data.Objects.Properties.Geometry.from_revit_conversion import (
    convert_bounding_box_to_flattened_2d_points,
    convert_xyz_in_data_geometry_polygons,
)

from duHast.Revit.Common.Geometry.solids import get_bounding_box_from_family_geometry


def populate_data_door_object(doc, revit_door):
    """
    Returns a custom ceiling data objects populated with some data from the revit model ceiling past in.

    - door id
    - door type name
    - door mark
    - door type mark
    - door level name
    - door level id
    - door offset from level

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_door: A revit door instance.
    :type revit_door: Autodesk.Revit.DB.FamilyInstance

    :return: A data door object instance.
    :rtype: :class:`.DataDoor`
    """

    # set up data class object
    data_door = dDoor.DataDoor()
    # get door bounding box
    # set a default option
    opts = Options()

    door_bounding_box = get_bounding_box_from_family_geometry(
        revit_door.get_Geometry(opts)
    )
    # only export door data if 3D geometry is available
    # if no geometry is available, return None
    # some families of type door do not have any 3D geometry... ignore those for now
    if (isinstance(door_bounding_box, BoundingBoxXYZ) == False):
        return None
    revit_door_data_geometry = convert_bounding_box_to_flattened_2d_points(
        door_bounding_box
    )
    
    if len(revit_door_data_geometry.outer_loop) > 0:
        door_point_groups_as_doubles = []
        data_geo_converted = convert_xyz_in_data_geometry_polygons(
            doc, revit_door_data_geometry
        )
        door_point_groups_as_doubles.append(data_geo_converted)
        data_door.polygon = door_point_groups_as_doubles
        # get design set data
        design_set_data = rDesignO.get_design_set_option_info(doc, revit_door)
        data_door.design_set_and_option.option_name = design_set_data[
            "designOptionName"
        ]
        data_door.design_set_and_option.set_name = design_set_data["designSetName"]
        data_door.design_set_and_option.is_primary = design_set_data["isPrimary"]

        # get type properties
        data_door.type_properties.id = revit_door.GetTypeId().IntegerValue
        data_door.type_properties.name = Element.Name.GetValue(revit_door).encode(
            "utf-8"
        )
        ceiling_type = doc.GetElement(revit_door.GetTypeId())

        # custom parameter value getters
        value_getter = {
            StorageType.Double: rParaGet.getter_double_as_double_converted_to_metric,
            StorageType.Integer: rParaGet.getter_int_as_int,
            StorageType.String: rParaGet.getter_string_as_UTF8_string,  # encode ass utf 8 just in case
            StorageType.ElementId: rParaGet.getter_element_id_as_element_int,  # needs to be an integer for JSON encoding
            str(None): rParaGet.getter_none,
        }
        data_door.type_properties.properties = (
            rParaGet.get_all_parameters_and_values_wit_custom_getters(
                ceiling_type, value_getter
            )
        )

        # get instance properties
        data_door.instance_properties.id = revit_door.Id.IntegerValue
        data_door.instance_properties.properties = (
            rParaGet.get_all_parameters_and_values_wit_custom_getters(
                revit_door, value_getter
            )
        )

        # get level properties
        data_door.level.name = Element.Name.GetValue(
            doc.GetElement(revit_door.LevelId)
        ).encode("utf-8")
        data_door.level.id = revit_door.LevelId.IntegerValue
        data_door.level.offset_from_level = rParaGet.get_built_in_parameter_value(
            revit_door, BuiltInParameter.CEILING_HEIGHTABOVELEVEL_PARAM
        )  # offset from level

        # get the model name
        if doc.IsDetached:
            data_door.revit_model.name = "Detached Model"
        else:
            data_door.revit_model.name = doc.Title

        # get phasing information
        data_door.phasing.created = rPhase.get_phase_name_by_id(
            doc,
            rParaGet.get_built_in_parameter_value(
                revit_door,
                BuiltInParameter.PHASE_CREATED,
                rParaGet.get_parameter_value_as_element_id,
            ),
        ).encode("utf-8")
        data_door.phasing.demolished = rPhase.get_phase_name_by_id(
            doc,
            rParaGet.get_built_in_parameter_value(
                revit_door,
                BuiltInParameter.PHASE_DEMOLISHED,
                rParaGet.get_parameter_value_as_element_id,
            ),
        ).encode("utf-8")

        return data_door
    else:
        return None


def get_all_door_data(doc, filter_family_names = []):
    """
    Gets a list of door data objects for each door element in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of data door instances.
    :rtype: list of :class:`.DoorCeiling`
    """

    all_door_data = []
    doors = get_door_instances(doc)
    for door in doors:
        # check if door is in filter list
        if len(filter_family_names) > 0:
            if door.Symbol.FamilyName not in filter_family_names:
                continue
        door_data = populate_data_door_object(doc, door)
        if door_data is not None:
            all_door_data.append(door_data)
    return all_door_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit railings. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family import purge_unused_family_types as rFamPurge
from duHast.Revit.Common import purge_utils as rPurgeUtils
from duHast.Revit.Railings.balusters import (
    get_all_baluster_symbols_ids,
    get_baluster_types_from_railings,
)
from duHast.Revit.Railings.railings import (
    get_in_place_railing_type_ids_in_model,
    get_all_railing_type_ids_by_class_and_category,
)
from duHast.Revit.Railings.Utility.railing_family_names import (
    BUILTIN_RAILING_TYPE_FAMILY_NAMES,
)
from duHast.Revit.Railings.Utility.merge_lists import merge_into_unique_list
from duHast.Revit.Railings.Utility.railings_type_sorting import (
    sort_railing_types_by_family_name,
)


def get_used_railing_type_ids(doc):
    """
    Gets all used Railing element types available in model excluding in place types.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_railing_type_ids_by_class_and_category, 1
    )
    return ids


def family_no_types_in_use(fam_type_ids, un_used_type_ids):
    """
    Compares two lists of element ids and returns False if any element id in first list is not in the second list.
    Returns False if any symbols (types) of a family (first list) are in use in a model (second list).
    TODO: repetitive code...Consider generic function!
    :param fam_type_ids: List of family symbols (types).
    :type fam_type_ids: List of Autodesk.Revit.DB.ElementId
    :param un_used_type_ids: List of unused family symbols (types)
    :type un_used_type_ids: List of Autodesk.Revit.DB.ElementId
    :return: True if all ids in first list are also in second list, otherwise False.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in un_used_type_ids:
            match = False
            break
    return match


def get_unused_non_in_place_railing_type_ids_to_purge(doc):
    """
    Gets all unused Railing type ids for:
    - Top Rail
    - Rail support
    - Hand rail
    - Rail termination
    - Railing Systems
    - loaded families
    Excludes any in place family types.
    This method can be used to safely delete unused railing types:
    In the case that no railing instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one railing type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    # get unused type ids
    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_railing_type_ids_by_class_and_category, 0
    )
    # make sure there is at least on Railing type per system family left in model
    railing_types = sort_railing_types_by_family_name(doc)
    for key, value in railing_types.items():
        if key in BUILTIN_RAILING_TYPE_FAMILY_NAMES:
            if family_no_types_in_use(value, ids) == True:
                # remove one type of this system family from unused list
                ids.remove(value[0])
    return ids


def get_used_in_place_railing_type_ids(doc):
    """
    Gets all used in place railing type ids.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of in place railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_in_place_railing_type_ids_in_model, 1
    )
    return ids


def get_unused_in_place_railing_type_ids(doc):
    """
    Gets all unused in place railing type ids.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of in place railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, get_in_place_railing_type_ids_in_model, 0
    )
    return ids


def get_unused_in_place_railing_ids_for_purge(doc):
    """
    Gets symbol(type) ids and family ids (when no type is in use) of in place Railing families which can be purged.
    This method can be used to safely delete unused in place railing types and families.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of in place railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_in_place_railing_type_ids
    )
    return ids


def get_used_baluster_type_ids(doc):
    """
    Gets all used baluster type ids in the model.
    Used: at least one instance of this family symbol (type) is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used_in_model = rPurgeUtils.get_used_unused_type_ids(
        doc, get_all_baluster_symbols_ids, 1
    )
    ids_used_in_railings = get_baluster_types_from_railings(doc)
    ids = merge_into_unique_list(ids, ids_used_in_model)
    ids = merge_into_unique_list(ids, ids_used_in_railings)
    return ids


def get_unused_baluster_type_ids(doc):
    """
    Gets all unused baluster type ids in the model.
    Unused: Not one instance of this family symbol (type) is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    ids_used = get_used_baluster_type_ids(doc)
    ids_available = get_all_baluster_symbols_ids(doc)
    for id in ids_available:
        if id not in ids_used:
            ids.append(id)
    return ids


def get_unused_baluster_type_ids_for_purge(doc):
    """
    Gets all unused baluster type ids in the model.
    Unused: at least one instance of this family symbol (type) is placed in the model.
    This method can be used to safely delete unused baluster families and symbols.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(doc, get_unused_baluster_type_ids)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data storage base class used for geometry aspects of Revit elements.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- contains 

    - polygon
    - topology cell (WIP)

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Utilities.Objects import base

from duHast.Data.Objects.Properties.Geometry import geometry_polygon
from duHast.Data.Objects.Properties.Geometry import geometry_topo_cell


class DataElementGeometryBase(base.Base):
    def __init__(self, j, **kwargs):
        """
        Class constructor

        :param j: Json formatted string or dictionary
        :type j: str or dic

        :raises ValueError: 'Argument supplied must be of type string or type dictionary'
        """

        # ini super class to allow multi inheritance in children!
        # forwards all unused arguments
        super(DataElementGeometryBase, self).__init__(**kwargs)
        # check valid j input
        if j != None and len(j) > 0:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # check for polygon data
            # this is stored as a list since there could be multiple polygons representing an object
            geometry_data_list = []
            if geometry_polygon.DataPolygon.data_type in j:
                for item in j[geometry_polygon.DataPolygon.data_type]:
                    if "data_type" in item:
                        if item["data_type"]:
                            dummy = geometry_polygon.DataPolygon(item)
                            geometry_data_list.append(dummy)
                    else:
                        print("no data type in item")
            self.polygon = geometry_data_list

            # check for topo cell data
            if geometry_topo_cell.DataTopologyCell.data_type in j:
                self.topologic_cell = geometry_topo_cell.DataTopologyCell(
                    j[geometry_topo_cell.DataTopologyCell.DataType]
                )
            else:
                self.topologic_cell = geometry_topo_cell.DataTopologyCell()

        else:
            # initialise classes with default values
            self.polygon = []
            self.topologic_cell = geometry_topo_cell.DataTopologyCell()



import clr, os, sys, getopt

clr.AddReferenceByPartialName("PresentationCore")
clr.AddReferenceByPartialName("PresentationFramework")
clr.AddReferenceByPartialName("WindowsBase")
clr.AddReferenceByPartialName("IronPython")
clr.AddReferenceByPartialName("Microsoft.Scripting")

from System.Windows import Application

# define path to duHast
# to get to the root folder of this repo if this script is called directly from a powershell, batch or similar
ROOT_REPO_DIRECTORY = os.path.join(
    os.path.realpath(__file__), os.pardir, os.pardir, os.pardir, os.pardir
)
sys.path.insert(0,ROOT_REPO_DIRECTORY)


# import UI helper
from duHast.UI import file_list as fl
from duHast.UI.Objects.file_select_settings import FileSelectionSettings
from duHast.UI.file_list import get_revit_files_for_processing
from duHast.UI.Objects.XamlLoader import XamlLoader
from duHast.UI import workloader as wl

from duHast.Utilities.files_io import file_exist
from duHast.Utilities.directory_io import directory_exists
from duHast.Utilities.console_out import output_with_time_stamp

# xaml helper classes
from ViewModel import ViewModel

# Define the path to your XAML file
CURRENT_SCRIPT_DIRECTORY = os.path.dirname(__file__)
XAML_FILE = "ui_v2.xaml"  # Replace with your XAML file name
XAML_FULL_FILE_NAME = os.path.join(CURRENT_SCRIPT_DIRECTORY, XAML_FILE)


def main(argv):
    """
    Entry point.

    :param argv: A list of string representing arguments past in.
    :type argv: [str]
    """

    # get arguments
    got_args, settings = process_args(argv)
    if got_args:
        # retrieve revit file data
        revit_files = get_revit_files_for_processing(
            settings.input_directory,
            settings.incl_sub_dirs,
            settings.revit_file_extension,
        )
        # check whether this is a BIM360 project or file system and assign
        # data retriever method accordingly
        if is_bim_360_file(revit_files):
            get_data = fl.bucket_to_task_list_bim_360
        else:
            get_data = fl.bucket_to_task_list_file_system
        
        # check if anything came back
        if len(revit_files) > 0:
            # load xaml
            xaml = XamlLoader(XAML_FULL_FILE_NAME)
            # forms window
            window = xaml.Root
            
            # view model initialise
            view_model = ViewModel(
                window,
                settings.input_directory,
                settings.output_dir,
                settings.revit_file_extension,
                settings.output_file_num,
                settings.incl_sub_dirs,
                settings.filters,
                settings.filter_type,
            )
            # assign view model to xaml
            xaml.Root.DataContext = view_model
            
            # show application
            Application().Run(xaml.Root)
            
            # debugging
            '''
            print("Destination Path {}".format(xaml.Root.DataContext.destination_path))
            print("Source Path {}".format(xaml.Root.DataContext.source_path))
            print(
                "Source File Extension {}".format(
                    xaml.Root.DataContext.source_file_extension
                )
            )
            print(
                "Include sub dirs {}\n".format(xaml.Root.DataContext.include_sub_dirs)
            )
            print("selected files {}\n".format(len(view_model.selected_files)))
            print("filter {}\n".format(view_model.filter_text))
            for rf in view_model.revit_files:
                print("file {}".format(rf.name))
            for frf in view_model.filtered_revit_files:
                print("filtered files: {}".format(frf.name))
            print("debug {}".format("\n".join(view_model.debug)))
            '''
            
            # get the dialog result (ok vs cancel buttons)
            ui_result = view_model.DialogResult
            if ui_result:
                # build bucket list
                buckets = wl.distribute_workload(
                    settings.output_file_num,
                    view_model.selected_files,
                    fl.get_file_size,
                )
                # write out file lists
                counter = 0
                for bucket in buckets:
                    file_name = os.path.join(
                        settings.output_dir, "Tasklist_" + str(counter) + ".txt"
                    )
                    status_write = fl.write_revit_task_file(file_name, bucket, get_data)
                    output_with_time_stamp(status_write.message)
                    counter += 1
                output_with_time_stamp("Finished writing out task files")
                sys.exit(0)
            else:
                # do nothing...
                output_with_time_stamp("No files selected!")
                sys.exit(2)
        else:
            # show message box
            output_with_time_stamp(
                "No files found matching extension: {} in location: {}!".format(
                    settings.revit_file_extension, settings.input_directory
                )
            )
            sys.exit(2)
    else:
        # invalid or no args provided... get out
        sys.exit(1)


def process_args(argv):
    """
    Processes past in arguments and checks whether inputs are valid.

    :param argv: List of arguments
    :type argv: _type_

    :return:
        - True if arguments past in are valid, otherwise False.
        - FIle select settings object instance.
    :rtype: bool, :class:`.FileSelectionSettings`
    """

    input_dir_file = ""
    output_directory = ""
    output_file_number = 1
    revit_file_extension = ".rvt"
    include_sub_dirs_in_search = False
    got_args = False
    try:
        opts, args = getopt.getopt(
            argv,
            "hsxi:o:n:e:f:",
            ["subDir", "TextFilterIsOR", "inputDirectory=", "outputDirectory=", "numberOfOutputFiles=", "fileExtension=", "filterText="],
        )
    except getopt.GetoptError as e:
        output_with_time_stamp(
            "file_select_ui.py -s -x <TextFilterIsOR> -i <inputDirectory> -o <outputDirectory> -n <numberOfOutputFiles> -e <fileExtension> -f <filterText> failed with exception: {}".format(
                e
            )
        )
        
    # set a default value (True) for the filter type
    filter_is_and = True
    # set default value for text filters
    filter_rules = []
    
    # check what args have been provided
    for opt, arg in opts:
        if opt == "-h":
            output_with_time_stamp(
                "file_select_ui.py -s <include sub dir> -x <TextFilterIsOR> -i <inputPath> -o <outputDirectory> -n <numberOfOutputFiles> -e <fileExtension> -f <filterText>"
            )
        elif opt in ("-s", "--subDir"):
            include_sub_dirs_in_search = True
        elif opt in ("-x", "--TextFilterIsOR"):
            filter_is_and = False
        elif opt in ("-i", "--inputDirectory"):
            input_dir_file = arg
            got_args = True
        elif opt in ("-o", "--outputDirectory"):
            output_directory = arg
            got_args = True
        elif opt in ("-n", "--numberOfOutputFiles"):
            try:
                value = int(arg)
                output_file_number = value
                got_args = True
            except ValueError:
                output_with_time_stamp("{}: value is not an integer".format(arg))
                got_args = False
        elif opt in ("-e", "--fileExtension"):
            revit_file_extension = arg.strip()
            got_args = True
        elif opt in ("-f", "--filterText"):
            filter_text = arg.strip()
            filter_rules = [rule.strip() for rule in filter_text.split(";")]

    # check if input values are valid
    if output_file_number < 0 or output_file_number > 100:
        got_args = False
        output_with_time_stamp(
            "The number of output files must be bigger then 0 and smaller then 100"
        )
    
    # need to check if path provided is a directory or a file path
    if os.path.isdir(input_dir_file):
        if not directory_exists(input_dir_file):
            got_args = False
            output_with_time_stamp(
                "Invalid input directory or file path: {}".format(input_dir_file)
            )
    elif(os.path.isfile(input_dir_file)):
        if not file_exist(input_dir_file):
            got_args = False
            output_with_time_stamp(
                "Invalid input file path: {}".format(input_dir_file)
            )
    else:
        got_args = False
        output_with_time_stamp(
                "Invalid input path: {}".format(input_dir_file)
            )
    
    # check output directory
    if not directory_exists(output_directory):
        got_args = False
        output_with_time_stamp("Invalid output directory: {}".format(output_directory))
    if (
        revit_file_extension.lower() != ".rvt"
        and revit_file_extension.lower() != ".rfa"
    ):
        got_args = False
        output_with_time_stamp(
            "Invalid file extension: [{}] expecting: .rvt or .rfa".format(
                revit_file_extension
            )
        )
    
    return got_args, FileSelectionSettings(
        input_dir_file,
        include_sub_dirs_in_search,
        output_directory,
        output_file_number,
        revit_file_extension,
        filter_rules,
        filter_is_and
    )


def is_bim_360_file(revit_files):
    """
    Checks whether the first item in a file item list belongs to a BIM 360 project.

    Checks whether Project GUID property on file item object is None.

    :param revit_files: List of file items.
    :type revit_files: [:class:`.FileItem`]
    :return: True if BIM360 file, otherwise False.
    :rtype: bool
    """

    bim_360_file = False
    for revit_file in revit_files:
        if revit_file.bim_360_project_guid != None:
            bim_360_file = True
            break
    return bim_360_file


#: module entry
if __name__ == "__main__":
    main(sys.argv[1:])



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit generic annotation helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    FilteredElementCollector,
)

# --------------------------------------------- utility functions ------------------


# returns all  GenericAnnotation types in a model
# doc:   current model document
def get_all_generic_annotation_types_by_category(doc):
    """
    This will return a filtered element collector of all GenericAnnotation types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: _description_
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_GenericAnnotation)
        .WhereElementIsElementType()
    )
    return collector


# returns all  GenericAnnotation types in a model
# doc:   current model document
def get_all_generic_annotation_type_ids_by_category(doc):
    """
    This will return a list of all GenericAnnotation types (symbols) id's in the model excluding shared families.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: _description_
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = get_all_generic_annotation_types_by_category(doc)
    for c in col:
        parameter_match = False
        # get the family object to check whether it is a shared family
        fam = c.Family
        id = get_built_in_parameter_value(fam, BuiltInParameter.FAMILY_SHARED)
        if id != None:
            parameter_match = True
            if id == "No" and c.Id not in ids:
                ids.append(c.Id)
        if parameter_match == False:
            # family cant be of type shared...
            ids.append(c.Id)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families parameter helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library
# utility functions for most commonly used Revit API tasks
from duHast.Revit.Common import transaction as rTran

# class used for stats reporting
from duHast.Utilities.Objects import result as res

# import Autodesk Revit DataBase namespace
import Autodesk.Revit.DB as rdb


def set_family_parameter_value_by_storage_type(param_w, manager, value):
    return_value = res.Result()
    try:
        if param_w.StorageType == rdb.StorageType.Double:
            # THIS IS THE KEY:  Use SetValueString instead of Set.  Set requires your data to be in
            # whatever internal units of measure Revit uses. SetValueString expects your value to
            # be in whatever the current DisplayUnitType (units of measure) the document is set to
            # for the UnitType associated with the parameter.
            #
            # So SetValueString is basically how the Revit GUI works.
            manager.SetValueString(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        elif param_w.StorageType == rdb.StorageType.ElementId:
            manager.Set(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        elif param_w.StorageType == rdb.StorageType.Integer:
            manager.Set(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        elif param_w.StorageType == rdb.StorageType.String:
            manager.Set(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        else:
            return_value.update_sep(False, "Parameter storage type is not supported!")

    except Exception as e:
        return_value.update_sep(
            False, "failed to set parameter value with exception: {}".format(e)
        )
    return return_value


def set_family_parameter_value(doc, manager, fam_para, value):
    # get the parameter
    param_w = manager.get_Parameter(fam_para.Definition.Name)
    # set-up action to be executed in transaction
    def action():
        action_return_value = res.Result()
        try:
            # attempt to change parameter value
            action_return_value = set_family_parameter_value_by_storage_type(
                param_w, manager, value
            )
        except Exception as e:
            action_return_value.status = False
            action_return_value.message = (
                fam_para.Definition.Name
                + " : Failed to set parameter value: with exception: "
                + str(e)
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "Setting parameter value")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def set_parameter_formula(doc, manager, fam_para, formula):
    def action():
        action_return_value = res.Result()
        try:
            # set parameter formula
            manager.SetFormula(fam_para, formula)
            action_return_value.message = (
                fam_para.Definition.Name + " : parameter formulas successfully set."
            )
            action_return_value.result.append(fam_para)
        except Exception as e:
            action_return_value.status = False
            action_return_value.message = (
                fam_para.Definition.Name
                + " : Failed to set parameter formula: with exception: "
                + str(e)
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "Setting parameter formula")
    return_value = rTran.in_transaction(transaction, action)
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
An implementation of Failur Handling Configuration
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Defines how to process warnings in Revit by storing the configuration for the failure handling.:
- Rolling back on warnings
- Printing warnings
- Rolling back on errors
- Printing errors
- Setting forced modal handling ( refer to autodesk documentation for more information )
- Setting clear after rollback ( refer to autodesk documentation for more information )

An output function is also provided to allow for custom output handling.

An instance of this class is passed to the failure handling function to determine how to handle failures.
"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Peter Smith, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.console_out import output
from duHast.Utilities.Objects.base import Base


class FailureHandlingConfig(Base):
    def __init__(
        self,
        roll_back_on_warning=False,
        print_warnings=False,
        roll_back_on_error=False,
        print_errors=False,
        set_forced_modal_handling=True,
        set_clear_after_rollback=True,
        output_function=output,
        **kwargs
    ):

        """
        Constructor for the FailureHandlingConfig class.

        :param roll_back_on_warning: Roll back the transaction on warnings.
        :type roll_back_on_warning: bool
        :param print_warnings: Print warnings.
        :type print_warnings: bool
        :param roll_back_on_error: Roll back the transaction on errors.
        :type roll_back_on_error: bool
        :param print_errors: Print errors.
        :type print_errors: bool
        :param set_forced_modal_handling: Set forced modal handling. ( refer to autodesk documentation for more information )
        :type set_forced_modal_handling: bool
        :param set_clear_after_rollback: Set clear after rollback. ( refer to autodesk documentation for more information )
        :type set_clear_after_rollback: bool
        :param output_function: A function to handle output.
        :type output_function: function
        :param kwargs: Additional keyword arguments.
        """

        super(FailureHandlingConfig, self).__init__(**kwargs)

        self.roll_back_on_warning = roll_back_on_warning
        self.print_warnings = print_warnings
        self.roll_back_on_error = roll_back_on_error
        self.print_errors = print_errors
        self.allowable_failures = None
        self.set_forced_modal_handling = set_forced_modal_handling
        self.set_clear_after_rollback = set_clear_after_rollback
        self.output_function = output_function
